<?php

namespace Drupal\xnttdm\Plugin\ExternalEntities\StorageClient;

use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Entity\ContentEntityType;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Entity\EntityStorageInterface;
use Drupal\Core\Entity\EntityTypeBundleInfoInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Entity\Query\ConditionInterface;
use Drupal\Core\Entity\Query\QueryInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\Core\Messenger\MessengerTrait;
use Drupal\Core\Plugin\PluginFormInterface;
use Drupal\Core\StringTranslation\TranslationInterface;
use Drupal\Core\Utility\Token;
use Drupal\external_entities\Entity\ExternalEntityInterface;
use Drupal\external_entities\Plugin\PluginFormTrait;
use Drupal\external_entities\StorageClient\StorageClientBase;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * External entities storage client to use local content entities as "external".
 *
 * @StorageClient(
 *   id = "content",
 *   label = @Translation("Content Entity"),
 *   description = @Translation("Use local Drupal content entity as external entities.")
 * )
 */
class ContentEntityClient extends StorageClientBase implements PluginFormInterface {

  use PluginFormTrait;
  use MessengerTrait;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The entity type bundle info service.
   *
   * @var \Drupal\Core\Entity\EntityTypeBundleInfoInterface
   */
  protected $entityTypeBundleInfo;

  /**
   * The used Drupal content type machine name.
   *
   * @var string
   */
  protected $contentTypeId;

  /**
   * The used Drupal content type bundle or an empty string if not used.
   *
   * @var string
   */
  protected $bundle;

  /**
   * The used Drupal content id key.
   *
   * @var string
   */
  protected $contentIdKey;

  /**
   * The used Drupal content bundle key.
   *
   * @var string
   */
  protected $bundleKey;

  /**
   * The used Drupal content storage instance.
   *
   * @var string
   */
  protected $drupalContentStorage;

  /**
   * Constructs a content entity storage client object.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin_id for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\Core\StringTranslation\TranslationInterface $string_translation
   *   The string translation service.
   * @param \Drupal\Core\Logger\LoggerChannelFactoryInterface $logger_factory
   *   The logger channel factory.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\Core\Entity\EntityFieldManagerInterface $entity_field_manager
   *   The entity field manager service.
   * @param \Drupal\Core\Utility\Token $token_service
   *   The token service.
   * @param \Drupal\Core\Entity\EntityTypeBundleInfoInterface $entity_type_bundle_info
   *   The entity bundle info service.
   */
  public function __construct(
    array $configuration,
    string $plugin_id,
    $plugin_definition,
    TranslationInterface $string_translation,
    LoggerChannelFactoryInterface $logger_factory,
    EntityTypeManagerInterface $entity_type_manager,
    EntityFieldManagerInterface $entity_field_manager,
    Token $token_service,
    EntityTypeBundleInfoInterface $entity_type_bundle_info,
  ) {
    $this->entityTypeBundleInfo = $entity_type_bundle_info;
    parent::__construct($configuration, $plugin_id, $plugin_definition, $string_translation, $logger_factory, $entity_type_manager, $entity_field_manager, $token_service);
    $this->init();
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('string_translation'),
      $container->get('logger.factory'),
      $container->get('entity_type.manager'),
      $container->get('entity_field.manager'),
      $container->get('token'),
      $container->get('entity_type.bundle.info')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function setConfiguration(array $configuration) {
    parent::setConfiguration($configuration);
    $this->init();
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    return [
      'content_type' => '',
    ];
  }

  /**
   * Helper method to initialize content type id and bundle.
   */
  protected function init() {
    [$this->contentTypeId, $bundle] = explode(
      ':',
      ($this->getConfiguration()['content_type'] ?? '') . ':'
    );
    $this->bundle = ($bundle ?? '') ? $bundle : '';
    try {
      $entity_type = $this->entityTypeManager->getDefinition($this->contentTypeId);
      $this->contentIdKey = $entity_type->getKey('id');
    }
    catch (\Exception $e) {
      $this->contentIdKey = 'id';
    }
    if (!empty($this->bundle)) {
      try {
        $this->bundleKey = $entity_type->getKey('bundle');
      }
      catch (\Exception $e) {
        $this->bundleKey = $this->contentIdKey;
      }
    }
    else {
      $this->bundleKey = $this->contentIdKey;
    }
  }

  /**
   * Return used Drupal content type id.
   *
   * @return string
   *   Used Drupal content type id (machine name).
   */
  public function getDrupalContentTypeId() :string {
    return $this->contentTypeId;
  }

  /**
   * Return used bundle associated to the Drupal content type.
   *
   * @return string
   *   Used bundle (machine name) associated to the Drupal content type.
   */
  public function getDrupalBundle() :string {
    return $this->bundle;
  }

  /**
   * Return associated Drupal content storage.
   *
   * @return \Drupal\Core\Entity\EntityStorageInterface
   *   A storage instance.
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   *   Thrown if the entity type doesn't exist.
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   *   Thrown if the storage handler couldn't be loaded.
   */
  public function getDrupalContentStorage() :EntityStorageInterface {
    if (!isset($this->drupalContentStorage)) {
      $this->drupalContentStorage = $this->entityTypeManager->getStorage($this->getDrupalContentTypeId());
    }
    return $this->drupalContentStorage;
  }

  /**
   * Returns a new query instance on used Drupal content.
   *
   * @param string $conjunction
   *   (optional) The logical operator for the query, either:
   *   - AND: all of the conditions on the query need to match.
   *   - OR: at least one of the conditions on the query need to match.
   *
   * @return \Drupal\Core\Entity\Query\QueryInterface
   *   A query object.
   */
  public function getDrupalContentQuery($conjunction = 'AND') :QueryInterface {
    $query = $this
      ->getDrupalContentStorage()
      ->getQuery($conjunction)
      ->accessCheck(TRUE);
    $content_type_id = $this->getDrupalContentTypeId();
    // If we got a bundle, make sure we only retrieve matching contents.
    $bundle = $this->getDrupalBundle();
    if (!empty($bundle) && ($bundle != $content_type_id)) {
      $entity_type = $this->entityTypeManager->getDefinition($content_type_id);
      $query->condition($this->bundleKey, $bundle);
    }
    return $query;
  }

  /**
   * Retrieves entity types and their bundles.
   *
   * @return array
   *   An array of entity types and their bundles suitable for a form "select"
   *   element. Bundles are grouped under their content type.
   */
  protected function getEntityTypesWithBundles() :array {
    $options = [];

    $definitions = $this->entityTypeManager->getDefinitions();
    /** @var \Drupal\Core\Entity\EntityTypeInterface $definition */
    foreach ($definitions as $entity_type_id => $definition) {
      if ($definition instanceof ContentEntityType) {
        // Check bundles.
        $bundles = $this->entityTypeBundleInfo->getBundleInfo($entity_type_id);
        if (empty($bundles)
            || ((1 == count($bundles))
                && (array_keys($bundles)[0] == $entity_type_id))
        ) {
          $options[$entity_type_id] = $definition->getLabel();
        }
        else {
          // We have multiple bundles.
          $bundle_options = [];
          foreach ($bundles as $bundle_id => $bundle_def) {
            $bundle_options[$entity_type_id . ':' . $bundle_id] = $bundle_def['label'];
          }
          $options['' . $definition->getLabel()] = $bundle_options;
        }
      }
    }
    return $options;
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
    $dm_id = ($form['#attributes']['id'] ??= uniqid('dm', TRUE));

    $content_type_list = $this->getEntityTypesWithBundles();

    $form = NestedArray::mergeDeep(
      $form,
      [
        'content_type' => [
          '#type' => 'select',
          '#title' => $this->t('Content type'),
          '#description' => $this->t('Select the content type.'),
          '#options' => $content_type_list,
          '#default_value' => $form_state->getValue('content_type')
          ?? $this->configuration['content_type']
          ?? NULL,
          '#empty_value' => '',
          '#sort_options' => TRUE,
          '#required' => TRUE,
        ],
      ]
    );

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function loadMultiple(?array $ids = NULL) :array {
    $data = [];

    $content_storage = $this->getDrupalContentStorage();
    $entities = $content_storage->loadMultiple($ids) ?? [];
    foreach ($entities as $entity) {
      $id = $entity->id();
      $data[$id] = $entity->toArray();
      $data[$id]['_id'] ??= $data[$id][$this->contentIdKey][0]['value'];
    }
    return $data;
  }

  /**
   * {@inheritdoc}
   */
  public function save(ExternalEntityInterface $entity) :int {
    $result = 0;

    $content_storage = $this->getDrupalContentStorage();
    $id = $entity->id();
    if (!empty($id)) {
      $drupal_entity = $content_storage->load($id);
      if (!empty($drupal_entity)) {
        foreach ($entity->toRawData() as $field_name => $field_values) {
          if ($drupal_entity->hasField($field_name)) {
            $drupal_entity->set($field_name, $field_values);
          }
        }
        $drupal_entity->save();
      }
      $result = SAVED_UPDATED;
    }
    else {
      $raw_data = $entity->toRawData();
      if (!empty($this->getDrupalBundle())) {
        $raw_data[$this->bundleKey] = $this->getDrupalBundle();
      }
      $drupal_entity = $content_storage->create($raw_data);
      $drupal_entity->save();
      $entity->set('id', $drupal_entity->id());
      $result = SAVED_NEW;
    }

    return $result;
  }

  /**
   * {@inheritdoc}
   */
  public function delete(ExternalEntityInterface $entity) {
    $content_storage = $this->getDrupalContentStorage();
    $drupal_entity = $content_storage->load($entity->id());
    if (!empty($drupal_entity)) {
      $drupal_entity->delete();
    }
  }

  /**
   * {@inheritdoc}
   */
  public function querySource(
    array $parameters = [],
    array $sorts = [],
    ?int $start = NULL,
    ?int $length = NULL,
  ) :array {
    $data = [];
    $query = $this->getDrupalContentQuery();
    if (!empty($sorts)) {
      $query->sort($sorts);
    }
    if (isset($start) || isset($length)) {
      $start ??= 0;
      $query->range($start, $length);
    }
    $this->addConditions($query, $parameters);
    $ids = $query->execute();
    if (!empty($ids)) {
      $data = array_values($this->loadMultiple($ids));
    }
    return $data;
  }

  /**
   * {@inheritdoc}
   */
  public function countQuerySource(array $parameters = []) :int {
    $query = $this->getDrupalContentQuery();
    $this->addConditions($query, $parameters);
    $query->count();
    $count = $query->execute();
    return $count;
  }

  /**
   * {@inheritdoc}
   */
  public function transliterateDrupalFilters(
    array $parameters,
    array $context = [],
  ) :array {
    // All filters are Drupal filters and therefore supported.
    $trans_filters = ['source' => $parameters, 'drupal' => []];

    return $this->transliterateDrupalFiltersAlter(
      $trans_filters,
      $parameters,
      $context
    );
  }

  /**
   * Process external entities filter as Drupal conditions.
   *
   * @param \Drupal\Core\Entity\Query\QueryInterface $query
   *   Query instance.
   * @param array $filters
   *   Filters provided by external entities system.
   *   @see \Drupal\external_entities\Entity\Query\External\Condition::subcompile()
   * @param \Drupal\Core\Entity\Query\ConditionInterface|null $condition_group
   *   Optional condition used for sub-condition groups.
   */
  protected function addConditions(QueryInterface $query, array $filters, ?ConditionInterface $condition_group = NULL) {
    if (!isset($condition_group)) {
      $condition_group = $query;
    }
    foreach ($filters as $filter) {
      if (!is_array($filter)) {
        continue;
      }
      if (!empty($filter['conjunction'])) {
        if (0 === strcasecmp($filter['conjunction'], 'and')) {
          $condition_and = $query->andConditionGroup();
          $this->addConditions($query, $filter['conditions'], $condition_and);
          $condition_group->condition($condition_and);
        }
        elseif (0 === strcasecmp($filter['conjunction'], 'or')) {
          $condition_or = $query->orConditionGroup();
          $this->addConditions($query, $filter['conditions'], $condition_or);
          $condition_group->condition($condition_or);
        }
      }
      elseif (!empty($filter['field'])) {
        $condition_group->condition($filter['field'], $filter['value'], $filter['operator'] ?? NULL);
      }
    }
  }

}
